@@ -62,7 +62,9 @@ struct CachedDir
   GSList      *monitors;
 
   guint have_read_entries : 1;
-  guint no_monitor_support : 1;
+  guint deleted : 1;
+
+  guint references : 28;
 };
 
 struct CachedDirMonitor
@@ -99,32 +101,9 @@ cached_dir_new (const char *name)
   return dir;
 }
 
-static void
-cached_dir_clear_entries (CachedDir *dir)
-{
-  g_slist_foreach (dir->entries,
-                   (GFunc) desktop_entry_unref,
-                   NULL);
-  g_slist_free (dir->entries);
-  dir->entries = NULL;
-}
-
-static void
-cached_dir_clear_subdirs (CachedDir *dir)
-{
-  g_slist_foreach (dir->subdirs,
-                   (GFunc) cached_dir_free,
-                   NULL);
-  g_slist_free (dir->subdirs);
-  dir->subdirs = NULL;
-}
-
 static void
 cached_dir_free (CachedDir *dir)
 {
-  cached_dir_clear_entries (dir);
-  cached_dir_clear_subdirs (dir);
-
   if (dir->dir_monitor)
     {
       menu_monitor_remove_notify (dir->dir_monitor,
@@ -138,6 +117,18 @@ cached_dir_free (CachedDir *dir)
   g_slist_free (dir->monitors);
   dir->monitors = NULL;
 
+  g_slist_foreach (dir->entries,
+                   (GFunc) desktop_entry_unref,
+                   NULL);
+  g_slist_free (dir->entries);
+  dir->entries = NULL;
+
+  g_slist_foreach (dir->subdirs,
+                   (GFunc) cached_dir_free,
+                   NULL);
+  g_slist_free (dir->subdirs);
+  dir->subdirs = NULL;
+
   g_free (dir->name);
   g_free (dir);
 }
@@ -193,9 +184,6 @@ cached_dir_find_relative_path (CachedDir  *dir,
   i = 0;
   while (split[i] != NULL)
     {
-      if (*(split[i]) == '\0')
-        continue;
-
       if (split[i + 1] != NULL)
         {
           if ((dir = find_subdir (dir, split[i])) == NULL)
@@ -237,9 +225,6 @@ cached_dir_lookup (const char *canonical)
     {
       CachedDir *subdir;
 
-      if (*(split[i]) == '\0')
-        continue;
-
       if ((subdir = find_subdir (dir, split[i])) == NULL)
         {
           subdir = cached_dir_new (split[i]);
@@ -329,6 +314,21 @@ cached_dir_add_subdir (CachedDir  *dir,
                        const char *path)
 {
   CachedDir *subdir;
+  GSList    *tmp;
+
+  tmp = dir->subdirs;
+  while (tmp != NULL)
+    {
+      subdir = (CachedDir *) tmp->data;
+
+      if (strcmp (subdir->name, basename) == 0)
+        {
+	  subdir->deleted = FALSE;
+          return TRUE;
+        }
+
+      tmp = tmp->next;
+    }
 
   subdir = cached_dir_new (basename);
 
@@ -359,8 +359,14 @@ cached_dir_remove_subdir (CachedDir  *dir,
 
       if (strcmp (subdir->name, basename) == 0)
         {
-	  cached_dir_free (tmp->data);
-          dir->subdirs = g_slist_delete_link (dir->subdirs, tmp);
+	  subdir->deleted = TRUE;
+
+	  if (subdir->references == 0)
+	    {
+	      cached_dir_free (subdir);
+	      dir->subdirs = g_slist_delete_link (dir->subdirs, tmp);
+	    }
+
           return TRUE;
         }
 
@@ -400,23 +406,25 @@ handle_cached_dir_changed (MenuMonitor      *monitor,
 {
   gboolean  handled = FALSE;
   char     *basename;
+  char     *dirname;
 
-  menu_verbose ("Notified of '%s' %s - invalidating cache\n",
+  menu_verbose ("'%s' notified of '%s' %s - invalidating cache\n",
+		dir->name,
                 path,
                 event == MENU_MONITOR_EVENT_CREATED ? ("created") :
                 event == MENU_MONITOR_EVENT_DELETED ? ("deleted") : ("changed"));
 
+  dirname  = g_path_get_dirname  (path);
   basename = g_path_get_basename (path);
 
+  dir = cached_dir_lookup (dirname);
+
   if (g_str_has_suffix (basename, ".desktop") ||
       g_str_has_suffix (basename, ".directory"))
     {
       switch (event)
         {
         case MENU_MONITOR_EVENT_CREATED:
-          handled = cached_dir_add_entry (dir, basename, path);
-          break;
-
         case MENU_MONITOR_EVENT_CHANGED:
           handled = cached_dir_update_entry (dir, basename, path);
           break;
@@ -452,6 +460,7 @@ handle_cached_dir_changed (MenuMonitor      *monitor,
     }
 
   g_free (basename);
+  g_free (dirname);
 
   if (handled)
     {
@@ -497,7 +506,6 @@ cached_dir_load_entries_recursive (CachedDir  *dir,
       return FALSE;
     }
 
-  cached_dir_clear_entries (dir);
   cached_dir_ensure_monitor (dir, dirname);
 
   fullpath = g_string_new (dirname);
@@ -539,52 +547,6 @@ cached_dir_load_entries_recursive (CachedDir  *dir,
   return TRUE;
 }
 
-static char *
-cached_dir_get_full_path (CachedDir *dir)
-{
-  GString *str;
-  GSList  *parents = NULL;
-  GSList  *tmp;
-
-  if (dir == NULL || dir->parent == NULL)
-    return g_strdup ("/");
-
-  while (dir->parent != NULL)
-    {
-      parents = g_slist_prepend (parents, dir->name);
-      dir = dir->parent;
-    }
-
-  str = g_string_new ("/");
-
-  tmp = parents;
-  while (tmp != NULL)
-    {
-      g_string_append (str, tmp->data);
-      g_string_append_c (str, '/');
-
-      tmp = tmp->next;
-    }
-
-  g_slist_free (parents);
-
-  return g_string_free (str, FALSE);
-}
-
-static CachedDir *
-cached_dir_load (const char *canonical_path)
-{
-  CachedDir *retval;
-
-  menu_verbose ("Loading cached dir \"%s\"\n", canonical_path);
-
-  retval = cached_dir_lookup (canonical_path);
-
-  cached_dir_load_entries_recursive (retval, canonical_path);
-
-  return retval;
-}
-
 static void
 cached_dir_add_monitor (CachedDir                 *dir,
                         EntryDirectory            *ed,
@@ -645,32 +607,51 @@ cached_dir_remove_monitor (CachedDir                 *dir,
 }
 
 static void
-cached_dir_ensure_loaded (CachedDir *dir)
+cached_dir_add_reference (CachedDir *dir)
 {
-  char *path;
+  dir->references++;
 
-  if (dir->have_read_entries)
-    return;
-
-  path = cached_dir_get_full_path (dir);
-  cached_dir_load_entries_recursive (dir, path);
-  g_free (path);
+  if (dir->parent != NULL)
+    {
+      cached_dir_add_reference (dir->parent);
+    }
 }
 
-static GSList*
-cached_dir_get_subdirs (CachedDir *dir)
+static void
+cached_dir_remove_reference (CachedDir *dir)
 {
-  cached_dir_ensure_loaded (dir);
+  CachedDir *parent;
 
-  return dir->subdirs;
-}
+  parent = dir->parent;
 
-static GSList*
-cached_dir_get_entries (CachedDir   *dir)
-{
-  cached_dir_ensure_loaded (dir);
+  if (--dir->references == 0 && dir->deleted)
+    {
+      if (dir->parent != NULL)
+	{
+	  GSList *tmp;
 
-  return dir->entries;
+	  tmp = parent->subdirs;
+	  while (tmp != NULL)
+	    {
+	      CachedDir *subdir = tmp->data;
+
+	      if (!strcmp (subdir->name, dir->name))
+		{
+		  parent->subdirs = g_slist_delete_link (parent->subdirs, tmp);
+		  break;
+		}
+
+	      tmp = tmp->next;
+	    }
+	}
+
+      cached_dir_free (dir);
+    }
+
+  if (parent != NULL)
+    {
+      cached_dir_remove_reference (parent);
+    }
 }
 
 /*
@@ -684,7 +665,6 @@ entry_directory_new_full (DesktopEntryType  entry_type,
                           const char       *legacy_prefix)
 {
   EntryDirectory *ed;
-  CachedDir      *cd;
   char           *canonical;
 
   menu_verbose ("Loading entry directory \"%s\" (legacy %s)\n",
@@ -699,12 +679,14 @@ entry_directory_new_full (DesktopEntryType  entry_type,
       return NULL;
     }
 
-  cd = cached_dir_load (canonical);
-  g_assert (cd != NULL);
-
   ed = g_new0 (EntryDirectory, 1);
 
-  ed->dir           = cd;
+  ed->dir = cached_dir_lookup (canonical);
+  g_assert (ed->dir != NULL);
+
+  cached_dir_add_reference (ed->dir);
+  cached_dir_load_entries_recursive (ed->dir, canonical);
+
   ed->legacy_prefix = g_strdup (legacy_prefix);
   ed->entry_type    = entry_type;
   ed->is_legacy     = is_legacy != FALSE;
@@ -749,6 +731,8 @@ entry_directory_unref (EntryDirectory *ed)
 
   if (--ed->refcount == 0)
     {
+      cached_dir_remove_reference (ed->dir);
+
       ed->dir        = NULL;
       ed->entry_type = DESKTOP_ENTRY_INVALID;
       ed->is_legacy  = FALSE;
@@ -793,14 +777,15 @@ entry_directory_get_directory (EntryDirectory *ed,
 }
 
 static char *
-get_desktop_file_id_from_path (EntryDirectory *ed,
-			       const char     *relative_path)
+get_desktop_file_id_from_path (EntryDirectory   *ed,
+			       DesktopEntryType  entry_type,
+			       const char       *relative_path)
 {
   char *retval;
   
   retval = NULL;
 
-  if (ed->entry_type == DESKTOP_ENTRY_DESKTOP)
+  if (entry_type == DESKTOP_ENTRY_DESKTOP)
     {
       if (!ed->is_legacy)
 	{
@@ -848,9 +833,12 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
   GSList *tmp;
   int     relative_path_len;
 
+  if (cd->deleted)
+    return TRUE;
+
   relative_path_len = relative_path->len;
 
-  tmp = cached_dir_get_entries (cd);
+  tmp = cd->entries;
   while (tmp != NULL)
     {
       DesktopEntry *entry = tmp->data;
@@ -863,7 +851,9 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
           g_string_append (relative_path,
                            desktop_entry_get_basename (entry));
 
-	  file_id = get_desktop_file_id_from_path (ed, relative_path->str);
+	  file_id = get_desktop_file_id_from_path (ed,
+						   ed->entry_type,
+						   relative_path->str);
 
           ret = func (ed, entry, file_id, set, user_data);
 
@@ -878,7 +868,7 @@ entry_directory_foreach_recursive (EntryDirectory            *ed,
       tmp = tmp->next;
     }
 
-  tmp = cached_dir_get_subdirs (cd);
+  tmp = cd->subdirs;
   while (tmp != NULL)
     {
       CachedDir *subdir = tmp->data;
@@ -933,7 +923,7 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
   if (subdirs)
     *subdirs = NULL;
 
-  tmp = cached_dir_get_entries (ed->dir);
+  tmp = ed->dir->entries;
   while (tmp != NULL)
     {
       DesktopEntry *entry = tmp->data;
@@ -946,7 +936,9 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
         {
           char *file_id;
 
-          file_id = get_desktop_file_id_from_path (ed, basename);
+          file_id = get_desktop_file_id_from_path (ed,
+						   DESKTOP_ENTRY_DESKTOP,
+						   basename);
 
           desktop_entry_set_add_entry (desktop_entries,
                                        entry,
@@ -968,12 +960,15 @@ entry_directory_get_flat_contents (EntryDirectory   *ed,
 
   if (subdirs)
     {
-      tmp = cached_dir_get_subdirs (ed->dir);
+      tmp = ed->dir->subdirs;
       while (tmp != NULL)
         {
           CachedDir *cd = tmp->data;
 
-          *subdirs = g_slist_prepend (*subdirs, g_strdup (cd->name));
+	  if (!cd->deleted)
+	    {
+	      *subdirs = g_slist_prepend (*subdirs, g_strdup (cd->name));
+	    }
 
           tmp = tmp->next;
         }
